// Copyright (c) 2014 Intel Corporation
// All rights reserved.
//
// WARRANTY DISCLAIMER
//
// THESE MATERIALS ARE PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL INTEL OR ITS
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
// OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THESE
// MATERIALS, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// Intel Corporation is the author of the Materials, and requests that all
// problem reports or change requests be submitted to it directly

// The code in this file operates on single-byte three-channel pixels.
#define CHANNELS	3

// A function to write out diagnostics showing the work-group size at
// various interesting points.
void WriteDiagnostics(__global struct NonUniformDiagnostics* diags);

// The kernel that implements the box blur.  This kernel does not necessarily
// represent the best way to implement a box blur.  It is just used as
// illustration in this tutorial.  It operates on 3-color, single-byte-per-
// color data from the input buffer, putting the results in the output
// buffer.  Because box blur requires a 3x3 neighborhood of pixels, this
// kernel should be launched with a global size that has two less elements
// in each dimension (one for each edge). 
__kernel void BoxBlur(__global uchar* input, uint input_width,
                      __global uchar* output, uint output_width,
                      __global struct NonUniformDiagnostics* diags)
{
    // The coordinates of the offset index are offset by 1 to handle the 
    // border
    uint out_x = get_global_id(0) + 1;
    uint out_y = get_global_id(1) + 1;
    __global uchar* out_ptr = output + 
                              (out_y * output_width + out_x) * CHANNELS;

    // This is the pointer to the middle of the 3x3 region for the box blur
    __global uchar* in_ptr = input + (out_y * input_width + out_x) * CHANNELS;

    // Load the 3x3 region of input pixels and sum up the values.  We do this 
    // in a uint because we could overflow a uchar
    uint3 pixel;
    pixel  = convert_uint3(vload3(0, in_ptr - (input_width + 1) * CHANNELS));
    pixel += convert_uint3(vload3(0, in_ptr - (input_width) * CHANNELS));
    pixel += convert_uint3(vload3(0, in_ptr - (input_width - 1) * CHANNELS));
    pixel += convert_uint3(vload3(0, in_ptr - CHANNELS));
    pixel += convert_uint3(vload3(0, in_ptr));
    pixel += convert_uint3(vload3(0, in_ptr + CHANNELS));
    pixel += convert_uint3(vload3(0, in_ptr + (input_width - 1) * CHANNELS));
    pixel += convert_uint3(vload3(0, in_ptr + (input_width) * CHANNELS));
    pixel += convert_uint3(vload3(0, in_ptr + (input_width + 1) * CHANNELS));

    // Divide by 9 to produce the average value, and write out the result.
    vstore3(convert_uchar3(pixel/9), 0, out_ptr);

    WriteDiagnostics(diags);
}

void WriteDiagnostics(__global struct NonUniformDiagnostics* diags)
{
    uint x = get_global_id(0);
    uint y = get_global_id(1);
    uint right = get_global_size(0) - 1;
    uint bottom = get_global_size(1) - 1;

    bool corner = false;
    __global struct WorkItemInfo* info;

    if ((x == 0) && (y == 0))
    {
        corner = true;
        info = &(diags->top_left);
    }
    if ((x == right) && (y == 0))
    {
        corner = true;
        info = &(diags->top_right);
    }
    if ((x == 0) && (y == bottom))
    {
        corner = true;
        info = &(diags->bottom_left);
    }
    if ((x == right) && (y == bottom))
    {
        corner = true;
        info = &(diags->bottom_right);
    }

    if (corner)
    {
        info->global_x = x;
        info->global_y = y;

        // In OpenCL 1.2 and 2.0, get_local_size() returns the actual size of
        // the current workgroup.
        info->local_size_x = get_local_size(0);
        info->local_size_y = get_local_size(1);
        #if (__OPENCL_C_VERSION__ >= 200)
        // In OpenCL 2.0, get_enqueued_local_size() returns the local size that
        // was requested in clEnqueueNDRange.  It may be different than the
        // actual local size in the remainder work-groups.
        info->enqueued_local_size_x = get_enqueued_local_size(0);
        info->enqueued_local_size_y = get_enqueued_local_size(1);
        #else
        // In OpenCL 1.2, get_enqueued_local_size() doesn't exist.
        info->enqueued_local_size_x = 0;
        info->enqueued_local_size_y = 0;
        #endif
    }
}
               