// Copyright (c) 2014 Intel Corporation
// All rights reserved.
//
// WARRANTY DISCLAIMER
//
// THESE MATERIALS ARE PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL INTEL OR ITS
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
// OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THESE
// MATERIALS, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// Intel Corporation is the author of the Materials, and requests that all
// problem reports or change requests be submitted to it directly

#ifndef IMAGE_H
#define IMAGE_H

#include <vector>
#include <string>
#include <fstream>
#include <memory>
#include <Windows.h>

#define __CL_ENABLE_EXCEPTIONS
#include <CL/cl.hpp>

// This is a very simple utility class for passing around images.  It can read
// and write 24-bit BMP files, and it allocates its backing data in a memory
// region aligned appropriately for zero-copy use in OpenCL on Intel Processor
// Graphics devices.

class Image
{
public:
    Image() : m_bytes(NULL, _aligned_free) {}

    // This constructor loads a bitmap from a file whose name is specified
    // in the argument.
    Image(const std::string& filename) : m_bytes(NULL, _aligned_free), 
                                         m_size(0)
    {
        try
        {
            std::ifstream input(filename, std::ios::binary);
            input.exceptions(std::ios::badbit | std::ios::failbit | 
                             std::ios::eofbit);

            // Load bitmap headers
            BITMAPFILEHEADER header = { NULL };
            input.read((char*)&header, sizeof(header));
            if ((header.bfType & 0xff) != 'B' || 
                (header.bfType >> 8) != 'M')
            {
                throw std::exception("File is not a bitmap!");
            }
            BITMAPINFOHEADER info_header = { NULL };
            input.read((char*)&info_header, sizeof(info_header));

            // Validate that this is a bitmap format that we can handle
            if (info_header.biSize < sizeof(info_header) || 
                info_header.biPlanes != 1 || 
                info_header.biBitCount != 24 || 
                info_header.biCompression != BI_RGB)
            {
                throw std::exception("File is an unsupported bitmap type!");
            }

            // Extract the dimensions from the image header
            m_width = info_header.biWidth;
            m_height = info_header.biHeight;

            AllocateMem();

            // Find the actual pixel data and load it in
            input.seekg(header.bfOffBits);  // Although bits is in the name, it 
                                            // is bytes
            size_t data_size = m_width * m_height * NUM_COLORS;

            input.read(m_bytes.get(), data_size);
        }
        catch (std::exception&)
        {
            throw;
        }
        catch (...)
        {
            // Clean up the exception so it makes more sense outside of this
            // class.
            throw std::exception("Could not read input file!");
        }
    }

    // This constructor is for allocating a sized image with undefined 
    // contents.  For example, to be used as the output for a function.
    Image (size_t width, size_t height) : 
        m_width(width), m_height(height), m_bytes(NULL, _aligned_free), 
        m_size(0)
    {
        AllocateMem();
    }

    // The copy constructor and assignment operators perform deep copies of
    // the image data.
    Image(const Image& image) :
        m_width(image.m_width), m_height(image.m_height), 
        m_bytes(NULL, _aligned_free), m_size(0)

    {
        AllocateMem();
        memcpy(m_bytes.get(), image.m_bytes.get(), 
               m_width * m_height * NUM_COLORS);
    }
    Image& operator = (const Image& image)
    {
        m_width = image.m_width;
        m_height = image.m_height;
        AllocateMem();
        memcpy(m_bytes.get(), image.m_bytes.get(), 
               m_width * m_height * NUM_COLORS);
        return *this;
    }

    // This is a move constructor and a move assignment operator, new features 
    // in C++11.  It allows us to initialize the class with a temporary 
    // instance.  It then adopts the contents without extra memory allocations.
    Image(Image&& image) :
        m_width(image.m_width), m_height(image.m_height), 
        m_bytes(NULL, _aligned_free), m_size(0)
    {
        image.m_bytes.swap(m_bytes);
        image.m_height = 0;
        image.m_width = 0;
        image.m_size = 0;
    }
    Image& operator = (Image&& image)
    {
        m_width = image.m_width;
        m_height = image.m_height;
        m_size = image.m_size;
        image.m_bytes.swap(m_bytes);
        image.m_height = 0;
        image.m_width = 0;
        image.m_size = 0;
        return *this;
    }

    virtual ~Image() { }

    // Write the data to a BMP file.
    void Write(const std::string& filename) 
    {
        try
        {
            std::ofstream output(filename, std::ios::binary | std::ios::trunc);
            output.exceptions(std::ios::badbit | std::ios::failbit);

            size_t data_size = m_width * m_height * NUM_COLORS;

            // Set up the headers
            BITMAPFILEHEADER header = { NULL };
            BITMAPINFOHEADER info_header = { NULL };
            header.bfType = (WORD)'B' | ((WORD)'M' << 8);
            header.bfSize = sizeof(header) + sizeof(info_header) + data_size;
            header.bfOffBits = sizeof(header) + sizeof(info_header);
            info_header.biSize = sizeof(info_header);
            info_header.biWidth = m_width;
            info_header.biHeight = m_height;
            info_header.biPlanes = 1;
            info_header.biBitCount = 24;
            info_header.biCompression = BI_RGB;
            info_header.biXPelsPerMeter = 3780;
            info_header.biYPelsPerMeter = 3780;

            output.write((char*)&header, sizeof(header));
            output.write((char*)&info_header, sizeof(info_header));
            output.write(m_bytes.get(), data_size);
        }
        catch (...)
        {
            throw std::exception("Could not write output file!");
        }
    }

    // Width and height in units of pixels.
    size_t get_width() const { return m_width; }
    size_t get_height() const { return m_height; }

    // This is the size of the allocated memory.  It may differ from width *
    // height * NUM_COLORS if it was padded to allow zero copy on Intel
    // Processor Graphics devices supporting OpenCL.
    size_t get_size() const { return m_size; }

    // The data is stored as an array of pixels.  Each pixel is represented
    // as NUM_COLORS single-byte quantities, packed together.
    char* get_data() { return m_bytes.get(); }

    static const int NUM_COLORS = 3;

protected:

    void AllocateMem()
    {
        // Allocate memory for the data, make sure it's aligned properly
        // for usage in OpenCL.  Additionally, set it up to use the right
        // free function.
        size_t new_size = m_width * m_height * NUM_COLORS;
        new_size = (((new_size - 1) / CACHE_LINE_SIZE) + 1) * CACHE_LINE_SIZE;
        if (new_size != m_size)
        {
            m_bytes = MemPtr((char*)_aligned_malloc(new_size, PAGE_SIZE),
                             _aligned_free);
            m_size = new_size;
        }
        if (!m_bytes.get())
        {
            throw std::exception("Out of memory!");
        }
    }

    // A special version of unique_ptr that calls the right _aligned_free 
    // function.
    typedef std::unique_ptr<char, decltype(&::_aligned_free)> MemPtr;
    static const int PAGE_SIZE = 4096;
    static const int CACHE_LINE_SIZE = 64;
    MemPtr m_bytes;
    size_t m_width;
    size_t m_height;
    size_t m_size;
};

#endif // IMAGE_H