// Copyright (c) 2014 Intel Corporation
// All rights reserved.
//
// WARRANTY DISCLAIMER
//
// THESE MATERIALS ARE PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL INTEL OR ITS
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
// OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THESE
// MATERIALS, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// Intel Corporation is the author of the Materials, and requests that all
// problem reports or change requests be submitted to it directly

#include <iostream>
#include <string>
#include <iomanip>
#include <sstream>
#include <exception>
#define __CL_ENABLE_EXCEPTIONS
#include <CL/cl.hpp>

#include "Image.h"
#include "ProgramFromResource.h"
#include "Diagnostics.h"

using namespace std;

static string GetDeviceString(cl::Device device);
static bool SupportsOpenCL2(cl::Device device);
static cl::Device GetDevice();
static cl::Kernel GetKernel(cl::Device device, cl::Context context, 
                            string build_options = string());
static Image RunBlurKernel(cl::Context context, cl::CommandQueue queue, 
                           cl::Kernel kernel, cl::NDRange global_size,
                           cl::NDRange local_size, Image& input, 
                           bool diagnostics_enabled = false);
static void PrintDiagnostics(struct NonUniformDiagnostics* diags);


static void print_usage(string command)
{
    cout << "Usage: " << command 
         << " [-h] [-?] [-i input_prefix] [-o output_prefix]" << endl; 
    cout << "  input_prefix: Base name of the .bmp file to read "
         << "(default: input)." << endl;
    cout << "  output_prefix: Base name of the .bmp files to write "
         << "(default: output)." << endl;
    cout << "                 Up to four output files are written "
         << "(output_{0|1|2|3}.bmp)." << endl;
}

int main(int argc, const char** argv)
{
    string input_base = "input";
    string output_base = "output";

    // Parse the command-line options
    int index = 1;
    while (index < argc)
    {
        const char* arg = argv[index];
        if (!string(arg).compare("-h") || !string(arg).compare("-?"))
        {
            print_usage(argv[0]);
            return 0;
        }
        if (!string(arg).compare("-i"))
        {
            ++index;
            if (index >= argc)
            {
                print_usage(argv[0]);
                return -1;
            }
            input_base = argv[index];
        }
        if (!string(arg).compare("-o"))
        {
            ++index;
            if (index >= argc)
            {
                print_usage(argv[0]);
                return -1;
            }
            output_base = argv[index];
        }
        ++index;
    }

    string input_file = input_base + ".bmp";
    vector<string> output_files;
    output_files.push_back(output_base + "_" + to_string(0) + ".bmp");
    output_files.push_back(output_base + "_" + to_string(1) + ".bmp");
    output_files.push_back(output_base + "_" + to_string(2) + ".bmp");
    output_files.push_back(output_base + "_" + to_string(3) + ".bmp");

    cout << "Input file: " << input_file << endl;
    cout << "Output files: " << output_files[0] << ", "
                             << output_files[1] << ", "
                             << output_files[2] << ", "
                             << output_files[3] << endl << endl;

    try
    {
        // Load the input image
        Image input(input_file);

        // Get the right device to use
        cl::Device device = GetDevice();

        // Print the device string
        cout << GetDeviceString(device) << endl << endl;

        // Get a context and command-queue using this device
        cl::Context context = cl::Context(device);
        cl::CommandQueue queue = cl::CommandQueue(context, device);

        // Get the box blur kernel compiled using OpenCL 1.2 (which is the 
        // default compilation, even on an OpenCL 2.0 device).  This allows
        // the code to show the pre-OpenCL 2.0 behavior.
        cl::Kernel kernel_1_2 = GetKernel(device, context);

        cl::Kernel kernel_2_0;
        if (SupportsOpenCL2(device))
        {
            // Get the box blur kernel compiled using OpenCL 2.0.  OpenCL 2.0
            // is required in order to use the non-uniform work-groups feature.
            kernel_2_0 = GetKernel(device, context, "-cl-std=CL2.0");
        }

        Image output;

        // Set the size of the global NDRange, to be used in all NDRange cases.
        // Since this is a box blur, we use a global size that is two elements 
        // smaller in each dimension.  This creates a range which often doesn't 
        // divide nicely by local work sizes we might commonly pick for running 
        // kernels.
        cl::NDRange global_size = cl::NDRange(input.get_width() - 2, 
                                              input.get_height() - 2);

        // Blur the image with a NULL local range using the OpenCL 1.2 compiled
        // kernel.
        cout << "Compiled with OpenCL 1.2 and using a NULL local size:" 
             << endl << endl;
        output = RunBlurKernel(context, queue, kernel_1_2, global_size, 
                               cl::NullRange, input, true);
        cout << endl;
        output.Write(output_files[0]);

        // Blur the image with an even local range using the OpenCL 1.2 
        // compiled kernel.  This won't work, even if we are running on an 
        // OpenCL 2.0 implementation.  The kernel has to be explicitly compiled
        // with OpenCL 2.0 compilation enabled in the compiler switches.
        try        
        {
            cout << "Compiled with OpenCL 1.2 and using an even local size:" 
                 << endl << endl;
            output = RunBlurKernel(context, queue, kernel_1_2, 
                                   global_size, cl::NDRange(16, 16), input, 
                                   true);
            cout << endl;
            output.Write(output_files[1]);
        }
        catch (...)
        {
            cout << "Trying to launch a non-uniform workgroup with a kernel "
                    "compiled using" << endl << 
                    "OpenCL 1.2 failed (as expected.)" << endl << endl;
        }

        if (SupportsOpenCL2(device))
        {
            // Blur the image with a NULL local range using the OpenCL 2.0 
            // compiled kernel.
            cout << "Compiled with OpenCL 2.0 and using a NULL local size:" 
                 << endl << endl;
            output = RunBlurKernel(context, queue, kernel_2_0, global_size, 
                                   cl::NullRange, input, true);
            cout << endl;
            output.Write(output_files[2]);

            // Blur the image with an even local range using the OpenCL 2.0 
            // compiled kernel.  This will only work on an OpenCL 2.0 device 
            // and compiler.
            cout << "Compiled with OpenCL 2.0 and using an even local size:" 
                 << endl << endl;
            output = RunBlurKernel(context, queue, kernel_2_0, 
                                   global_size, cl::NDRange(16, 16), input, 
                                   true);
            cout << endl;
            output.Write(output_files[3]);
        }
        else
        {
            cout << "This device does not support OpenCL 2.0, so non-uniform "
                    "work-groups cannot be " << endl << "used." << endl 
                 << endl;
        }
    }
    catch (cl::Error& e)
    {
        cerr << "Error: " << e.what() << " failed with error code " << 
             e.err() << endl;
        return -1;
    }
    catch (exception& e)
    {
        cerr << "Error: " << e.what() << endl;
        return -1;
    }

    cout << "Done!" << std::endl;
    return 0;
}

string GetDeviceString(cl::Device device)
{
    string device_string;
    device_string += string("Device: ") + 
                     device.getInfo<CL_DEVICE_NAME>() + 
                     string("\nVendor: ") +
                     device.getInfo<CL_DEVICE_VENDOR>();
    return device_string;
}

bool SupportsOpenCL2(cl::Device device)
{
    // If the implementation supports OpenCL 2.0, it supports non-uniform 
    // workgroups.  Check the device and the compiler for this.
    string dev_version = device.getInfo<CL_DEVICE_VERSION>();
    // The format of the version string is defined in the spec as
    // "OpenCL <major>.<minor> <vendor-specific>"
    dev_version = dev_version.substr(string("OpenCL ").length());
    dev_version = dev_version.substr(0, dev_version.find('.'));

    string lang_version = device.getInfo<CL_DEVICE_OPENCL_C_VERSION>();
    // The format of the version string is defined in the spec as
    // "OpenCL C <major>.<minor> <vendor-specific>"
    lang_version = lang_version.substr(string("OpenCL C ").length());
    lang_version = lang_version.substr(0, lang_version.find('.'));

    return (stoi(dev_version) >= 2 && stoi(lang_version) >= 2);
}

cl::Device GetDevice()
{
    // Find the GPU device if there is one.  Otherwise, use the CPU.
    cl::Platform platform = platform.getDefault();
    vector<cl::Device> devices;
    try
    {
        platform.getDevices(CL_DEVICE_TYPE_GPU, &devices);
    }
    catch (cl::Error)
    {
        platform.getDevices(CL_DEVICE_TYPE_CPU, &devices);
    }

    return devices[0];
}

cl::Kernel GetKernel(cl::Device device, cl::Context context, 
                     string build_options)
{
    // Get the program from a Windows resource so we don't have to hunt the
    // file system for it.
    vector<string> resources;
    resources.push_back("IDR_OPENCL_DIAGNOSTICS_H");
    resources.push_back("IDR_OPENCL_BLUR_C");
    ProgramFromResource program(context, resources.begin(), resources.end());

    // Build the program
    try
    {
        program.build(build_options.c_str());
    }
    catch (cl::Error)
    {
        std::string build_error = 
            program.getBuildInfo<CL_PROGRAM_BUILD_LOG>(device);
        OutputDebugStringA("Build log: \n");
        OutputDebugStringA(build_error.c_str());
        throw;
    }
    return cl::Kernel(program, "BoxBlur");
}

Image RunBlurKernel(cl::Context context, cl::CommandQueue queue, 
                    cl::Kernel kernel, cl::NDRange global_size,
                    cl::NDRange local_size, Image& input, 
                    bool diagnostics_enabled)
{
    // Create the output image with the same size as the input
    Image output(input.get_width(), input.get_height());

    // Create the buffers for input and output
    cl::Buffer input_buf(context, CL_MEM_USE_HOST_PTR, input.get_size(), 
                         input.get_data());
    cl::Buffer output_buf(context, CL_MEM_USE_HOST_PTR, output.get_size(),
                          output.get_data());

    // Set up a diagnostics buffer
    cl::Buffer diags_buf(context, CL_MEM_ALLOC_HOST_PTR, 
                         sizeof(struct NonUniformDiagnostics));

    // Set up the input and output arguments
    kernel.setArg(0, input_buf);
    kernel.setArg(1, (unsigned int)input.get_width());
    kernel.setArg(2, output_buf);
    kernel.setArg(3, (unsigned int)output.get_width());
    kernel.setArg(4, diags_buf);

    // Run the kernel on the input and output.  Since this is a box blur, we
    // use a global size that is two elements smaller in each dimension.  This
    // creates a range which often doesn't divide nicely by local work sizes
    // we might commonly pick for running kernels.
    queue.enqueueNDRangeKernel(kernel, cl::NullRange, global_size, local_size);

    // Synchronize the host data with the buffer
    queue.enqueueMapBuffer(output_buf, CL_TRUE, CL_MAP_READ, 0, 
                           output.get_size());
    queue.enqueueUnmapMemObject(output_buf, output.get_data());

    // Print diagnostics, if requested.  The diagnostics in this case show the
    // results of calling get_local_size() and get_enqueued_local_size() from
    // the work-items in each corner of the 2D NDRange.  This is just for 
    // illustration purposes, and probably wouldn't be done in production code.
    if (diagnostics_enabled)
    {
        struct NonUniformDiagnostics* diags = (struct NonUniformDiagnostics*)
            queue.enqueueMapBuffer(diags_buf, CL_TRUE, CL_MAP_READ, 0, 
                                   sizeof(struct NonUniformDiagnostics));
        PrintDiagnostics(diags);
        queue.enqueueUnmapMemObject(diags_buf, diags);
    }

    return output;
}

void PrintDiagnostics(struct NonUniformDiagnostics* diags)
{
    // Prints a table of global IDs, local sizes, and enqueued local sizes.
    string work_item_heading = "get_global_id()";
    string local_size_heading = "get_local_size()";
    string enqueued_local_size_heading = "get_enqueued_local_size()";
    string gap = "  ";
    auto PairToString = [] (unsigned int x, unsigned int y)
    {
        stringstream stream;
        stream << "(" << setw(3) << x << ", " << setw(3) << y << ")";
        return stream.str();
    };

    auto PrintWorkItemInfo = [=] (const struct WorkItemInfo& info)
    {
        cout << setw(work_item_heading.length()) 
             << PairToString(info.global_x, info.global_y) << gap;
        cout << setw(local_size_heading.length()) 
             << PairToString(info.local_size_x, info.local_size_y) << gap;

        // In the OpenCL 1.2 version of this kernel, the 
        // get_enqueued_local_size() call doesn't exist.  The kernel uses 0 to
        // indicate this case.
        string enqueued_local_size;
        if (!info.enqueued_local_size_x || !info.enqueued_local_size_y)
        {
            enqueued_local_size = "undefined";
        }
        else
        {
            enqueued_local_size = PairToString(info.enqueued_local_size_x, 
                                               info.enqueued_local_size_y);
        }
        cout << setw(enqueued_local_size_heading.length()) 
             << enqueued_local_size << gap;
        cout << endl;
    };

    const int first_width = 12;
    stringstream heading;
    heading << setw(first_width) << "Work Item" << gap 
            << work_item_heading << gap
            << local_size_heading;
    heading << gap << enqueued_local_size_heading;
    heading << endl;

    cout << heading.str();
    cout.fill('-');
    cout << setw(heading.str().length()) << "" << endl;
    cout.fill(' ');
    cout << setw(first_width) << "Top left" << gap;
    PrintWorkItemInfo(diags->top_left);
    cout << setw(first_width) << "Top right" << gap;
    PrintWorkItemInfo(diags->top_right);
    cout << setw(first_width) << "Bottom left" << gap;
    PrintWorkItemInfo(diags->bottom_left);
    cout << setw(first_width) << "Bottom right" << gap;
    PrintWorkItemInfo(diags->bottom_right);
    cout << endl;
}

